/*
 * Jobs service provides an obvservable (behaviorsubject) listing the jobs runnin on the current Identity
 * It uses tes.service to make the actual calls to get the job information.
 * The Joblist component holds the timer which repeats calls to this service.
 *
 */
import { Injectable } from '@angular/core';
import {Job} from './job';
import {Subscription, BehaviorSubject} from 'rxjs';
import {Identity} from './identity';
import {AuthorisationService} from './authorisation.service';
import { HttpClientModule, HttpClient, HttpHeaders, HttpErrorResponse } from '@angular/common/http';
import { BatchInterface } from './batchinterface';
import {Strudelapp} from './strudelapp';
import { BackendSelectionService } from './backend-selection.service';
import {NotificationsService} from './notifications.service';

@Injectable({
  providedIn: 'root'
})
export class JobsService {

  public joblist: BehaviorSubject<Job[]> = new BehaviorSubject([]);
  public statusMsg: BehaviorSubject<any>;
  private updateJobSub: Subscription;

constructor(private http: HttpClient, private authorisationService: AuthorisationService, private backendSelectionService: BackendSelectionService, private notifications: NotificationsService ) { 
}

  public setStatusMsg(statusMsg: BehaviorSubject<any>) {
    this.statusMsg = statusMsg;
  }

   private getJobsError(error: any, identity: Identity) {
     this.joblist.next([]);
     if (error.status == 0) {
       this.notifications.notify("A network error occurred. Please try again later");
       return
     }
     console.error(error);
     if (error.status == 401) {
       this.notifications.notify("Login expired. Please log in again.");
       this.authorisationService.updateAgentContents();
       return
     }
     if (error.status == 400) {
         if (error.error !== undefined && error.error.message !== undefined) {
          this.notifications.notify(error.error.message);
         }
         return
     }
   }

   getJobs(identity: Identity) {
    let headers = new HttpHeaders();
    let options = { headers: headers, withCredentials: true};
    // remove from the job list any jobs for identities that we don't know about
    let bi = new BatchInterface();
    bi.cancelcmd = identity.site.cancelcmd;
    bi.statcmd = identity.site.statcmd;
    let params = new URLSearchParams();
    params.set('statcmd',JSON.stringify(identity.site.statcmd));
    params.set('host',JSON.stringify(identity.site.host));
    params.set('username',JSON.stringify(identity.username));
    
    this.updateJobSub = this.http.get<Job[]>(this.backendSelectionService.apiserver.value.tes +'/stat'+'?'+params.toString(),options)
                  .subscribe(resp => this.updateJoblist(resp, identity),
                             error => this.getJobsError(error, identity));
  }

 updateJoblist(resp, identity: Identity) {
     // resp contains a javascript represnetiation of a list of jobs
     // We want to update the joblist BUT we don't want to create new Job objects
     // instead we want to reuse existing job objects removing any which are no longer valid
     // and adding any new ones. We also want the list sorted from largest jobid to smallest (oldest job)
     // The sort is lexographic since sometimes jobids are a string rather than a number
   
   var joblist: Job[] = []
   var jobquery: Job[] = <Job[]>resp;
   var lastjoblist: Job[] = this.joblist.value;
   var qjobids: any[] = [];
   var jobids: any[] = [];
   var j: Job;
   var newjob: Job;
   var idx: number;

   for (j of jobquery) {
       qjobids.push(j.jobid);
   }
   for (j of lastjoblist) {
       if (qjobids.indexOf(j.jobid) != -1) {
           if (jobids.indexOf(j.jobid) == -1) {
               idx = qjobids.indexOf(j.jobid)
               newjob = jobquery[idx]
               // These values in the job may change, but we need to keep using the old object
               j.state  = newjob.state;
               j.endtime = newjob.endtime;
               j.batch_host = newjob.batch_host;
               joblist.push(j);
               jobids.push(j.jobid);
           }
       }
   }
   for (j of jobquery) {
       if (jobids.indexOf(j.jobid) == -1) {
           joblist.push(j);
           jobids.push(j.jobid);
       }
   }
       
   for (j of joblist) {
     if (j.app === undefined || j.app == null) {
       j.app = this.getApp(j.appname,identity.site.appCatalog.value);
     }
     if (j.identity == undefined) {
        j.identity = identity;
     }
     if (j.connectionState == undefined) {
         j.connectionState = 0;
     }
   }
   joblist = joblist.sort((a,b) =>  (a.jobid < b.jobid)? 1:-1);
   this.joblist.next(joblist);
   //this.notifications.notify(null);
 }

  getApp(name: string,applist: any): any {

    var idx: number;
    var app: any;
    var sapp: Strudelapp;
    app = applist[0];
    for ( let item of applist) {
      sapp = <Strudelapp>item;
      let sname = sapp.name.toLowerCase().replace(/\s/g, '');
      let jname = name.toLowerCase().replace(/\s/g, '');
      idx = sname.indexOf(jname)
      if (idx == 0) {
        app = item;
        return app;
      }
      if (sapp.applist != null) {
        app = this.getApp(name,sapp.applist);
        if (app != null) {
          return app;
        }
      }
    }
    return null;
  }
}
