import { Injectable, EventEmitter } from '@angular/core';
import { HttpClientModule, HttpClient, HttpHeaders, HttpErrorResponse } from '@angular/common/http';
import { Observable, Subject, BehaviorSubject, of, from, interval } from 'rxjs';
import { fromEvent, throwError, Subscription, merge } from 'rxjs';
import { catchError, map, tap, filter } from 'rxjs/operators';
import { Job } from './job';
import { AppAction, Strudelapp, StrudelappInstance } from './strudelapp';
import { Computesite, Health } from './computesite';
import { Identity, AuthToken, KeyCert, SshAuthzServer } from './identity';
import { BatchInterface } from './batchinterface';
import { timer } from 'rxjs/observable/timer';
import { repeat, takeUntil, take } from 'rxjs/operators';
import { LocationStrategy, Location } from '@angular/common';
import { ActivatedRoute, Router } from '@angular/router';
import { AuthorisationService } from './authorisation.service';
import { environment } from '../environments/environment';
import { BackendSelectionService } from './backend-selection.service';
import {JobsService} from './jobs.service';
import {BrowserWindowService} from './browser-window.service';
import { NotificationsService } from './notifications.service';

/** The TES service contains ways to start Tunnels, and Execute programs
Its also responsible for querying a compute site for running jobs */



@Injectable({
  providedIn: 'root',
})
export class TesService {
public Base: string;
private twsproxy: string;
public jobs: any[];
private timerSubscription: any;
private appwindow: any;
private updateJobSub: Subscription;
private updateUserHealthSub: Subscription;
private cachetincidents: BehaviorSubject<Health[]>;
private nextUpdate: Subscription;
private cancelRequests$: Subject<boolean>;
public identitySubject: BehaviorSubject<Identity>;
public appSubject: BehaviorSubject<Strudelapp>;
public openWindow$: Subject<any>;


  constructor(private http: HttpClient,
              private authorisationService: AuthorisationService,
              private backendSelectionService: BackendSelectionService,
              private jobsService: JobsService,
              private location: Location,
              private browserWindow: BrowserWindowService,
              private notifications: NotificationsService,
  ) {
    this.cancelRequests$ = new Subject<boolean>();
    this.openWindow$ = new Subject<any>();
    this.backendSelectionService.apiserver.subscribe( (value) => { if (value != null) {this.twsproxy = value.tws ; this.Base = value.tes }});
 }



 private buildParams(app: Strudelapp, identity: Identity, batchinterface: BatchInterface, appinst?: any): string {
   let params = new URLSearchParams();
   let id = identity.copy_skip_catalog();
   id.site.appCatalog = null;
   if (appinst!== null)  {
     params.set('appinstance',JSON.stringify(appinst));
   }
   params.set('app',JSON.stringify(app));
   params.set('interface',JSON.stringify(batchinterface));
   params.set('identity',JSON.stringify(id));
   return params.toString();
 }



 private getBatchInterfaceError(error: any) {
   console.error(error);
 }

 public cancelHealthRequests() {
    this.cancelRequests$.next(true);
 }

 private getUserHealthError(error: any, identity: Identity) {
   identity.accountalerts.next([]);
   this.notifications.notify("There was an error checking your user account");
 }


getUserHealth(identity: Identity)  {
  let headers = new HttpHeaders();
  let options = { headers: headers, withCredentials: true};
  let params = new URLSearchParams();
  if (identity === undefined || identity === null) {
    return
  }
  if (identity.site.userhealth === undefined) {
    identity.accountalerts.next([]);
    return
  }
  params.set('statcmd',JSON.stringify(identity.site.userhealth));
  params.set('host',JSON.stringify(identity.site.host));
  params.set('username',JSON.stringify(identity.username));
  
  this.updateUserHealthSub = this.http.get<Health[]>(this.Base+'/stat'+'?'+params.toString(),options)
    .pipe(takeUntil(this.cancelRequests$))
    .subscribe(resp => this.addUserHealth(identity,resp), error => this.getUserHealthError(error,identity));
}

public getHealthAlerts(identity: Identity) {
  this.getCachetIncidents(identity);
  this.getUserHealth(identity);
}

public getCachetIncidents(identity: Identity) {
  let headers = new HttpHeaders();
  let options = { headers: headers, withCredentials: false};
  let params = new URLSearchParams();

  if (identity.site.cacheturis === undefined ||  identity.site.cacheturis.length == 0) {
    identity.systemalerts.next([]);
    return
  }
  for (let uri of identity.site.cacheturis) {
    this.http.get(uri,options)
      .pipe(takeUntil(this.cancelRequests$))
      .subscribe(resp => this.addCachetIncidents(identity,resp), error => this.getCachetIncidentsError(error,identity));
  }
}

private addCachetIncidents(identity,resp) {
  let ci = [];
  for (let i of resp.data) {
    if (i.status == 3 || i.status == 4) {
      continue;
    }
    let h = new Health();
    h.stat = 'error';
    h.msg = i.message;
    ci.push(h);
  }
  identity.systemalerts.next(ci);
}

private addUserHealth(identity,resp) {
  let ci = []
  for (let i of resp) {
    let h = new Health();
    h.stat = i.stat;
    h.msg = i.message;
    if (i.title != undefined) {
      h.title = i.title;
    }
    if (i.type != undefined) {
      h.type = i.type;
      h.data = i.data;
    }
    ci.push(h);
  }
  identity.accountalerts.next(ci);
}

 private getCachetIncidentsError(error: any, identity: Identity) {
   if (error.status == 0) {
     this.notifications.notify("A network error occurred. Please try again later");
     return
   }
   console.error(error);
   if (error.status == 401) {
     this.notifications.notify("Login expired. Please log in again.");
     this.authorisationService.updateAgentContents();
     return
   }
   if (error.status == 400) {
       if (error.error !== undefined && error.error.message !== undefined) {
        this.notifications.notify(error.error.message);
       }
       return
   }
 }


 submissionError(error: any) {
   if (error.status != 0) {
       if ('error' in error && 'message' in error.error) {
         this.notifications.notify(error.error.message);
       } else {
         this.notifications.notify('Job submission failed');
       }
   }
 }

 cancelError(error: any) {
     console.log('error canceling job',error);
   if (error.status != 0) {
       if ('error' in error && 'message' in error.error) {
         this.notifications.notify(error.error.message);
       } else {
         this.notifications.notify('Job Canceling failed');
       }
   }
 }

 buildBody(app: Strudelapp, appparams?: string) {
   return JSON.stringify({'app': app, 'appparams': appparams});
 }

 submit(app: Strudelapp, identity: Identity, batchinterface: BatchInterface, ftidentities?: Identity[], appparams?: any) {
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};
   this.notifications.notify('Submitting job');
   let paramstr = this.buildParams(app,identity,batchinterface);
   let keys = JSON.stringify(this.authorisationService.getKeys());
   let loggedin = JSON.stringify(this.authorisationService.loggedInAuthZ.value);
   let ids = [];
   if (ftidentities != undefined) {
     for (let id of ftidentities) {
         ids.push(id.copy_skip_catalog())
     }
   }
   let body = {'app': app, 'appparams': appparams, 'keys': keys, 'ids': JSON.stringify(JSON.stringify(ids))}
   this.http.post<any>(this.Base+'/submit'+'?'+paramstr, body, options)
                                          .subscribe(resp => { this.notifications.notify(null) },
                                                     error => this.submissionError(error));
 }

 submitted(resp: any, identity: Identity ) {
   this.jobsService.getJobs(identity);
 }

 cancel(job: Job) {
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};
   let data = {};
   let bi = new BatchInterface();
   bi.statcmd = job.identity.site.statcmd;
   bi.cancelcmd = job.identity.site.cancelcmd;
   let paramstr = this.buildParams(job.app,job.identity,bi);
   this.http.delete<any>(this.Base+'/cancel/'+job.jobid+'?'+paramstr, options)
         .subscribe(resp => this.submitted(resp,job.identity),
                    error => this.cancelError(error));
 }

 public getAppInstance(job: Job, action: AppAction) {
   let username = job.identity.username;
   let loginhost = job.identity.site.host;
   let batchhost = job.batch_host;
   let jobid = job.jobid;
   let params = new URLSearchParams;
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};
   params.set('cmd',JSON.stringify(action.paramscmd));
   let paramstr = params.toString();
   job.connectionState = 1;
   this.http.get<string>(this.Base+'/appinstance/'+username+'/'+loginhost+'/'+batchhost+'/'+jobid+'?'+paramstr, options)
     .subscribe(resp =>  { console.log('ran',action.paramscmd,'got',resp); job.appinst = resp; if (action.client != null) { this.createTunnel(job, action) } else { job.connectionState = 0} },
             error => { this.handleAppInstanceError(job,error) })
 }

 public createTunnel(job: Job, action: AppAction) {
   let username = job.identity.username;
   let loginhost = job.identity.site.host;
   let batchhost = job.batch_host;
   let params = new URLSearchParams;
   let headers = new HttpHeaders();
   job.connectionState = 2;

   if ((action.notunnel != undefined) && (action.notunnel == true)) {
     this.getAppUrl(job,action);
     return
   }
   let options = { headers: headers, withCredentials: true};
   let paramstr = params.toString();
   this.http.post<string>(this.Base+'/createtunnel/'+username+'/'+loginhost+'/'+batchhost+'?'+paramstr, job.appinst, options)
     .subscribe(() =>  { this.getAppUrl(job, action) }, error => {this.handleTunnelError(job,error)} )
 }

 public getAppUrl(job: Job, action: AppAction) {
   let params = new URLSearchParams;
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};
   let pseudoapp = {'client':{'redir':action.client.redir}};
   params.set('app',JSON.stringify(pseudoapp));
   params.set('appinst',JSON.stringify(job.appinst));
   let paramstr = params.toString();
   job.connectionState = 3;
   this.http.get<string>(this.Base+'/appurl?'+paramstr,options)
                .pipe(catchError(this.handleError))
     .subscribe((resp) => {  job.connectionState = 0; this.openWindow$.next( {'job':job, 'url': <string> resp, 'usebasicauth':action.client.usebasicauth})});
       //[this.browserWindowService.openAppWindow(resp,job,action)});
 }

 public getInterface(job: Job, action: AppAction) {
   this.getAppInstance(job, action);
 }


 public connect(job: Job, action: AppAction, appinst?: any) {
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};
   console.log('called connect');
   console.log(action);
   console.log(job);
   this.getInterface(job,action); // getInterface will subsequently called getAppInstance, which will call createTunnel, which will openAppWindow
 }

public runCommand(identity: Identity, command: string) {
  let headers = new HttpHeaders();
  let options = { headers: headers, withCredentials: true};
  // remove from the job list any jobs for identities that we don't know about
  let params = new URLSearchParams();
  if (identity === undefined || identity === null) {
    return
  }
  params.set('statcmd',JSON.stringify(command));
  params.set('host',JSON.stringify(identity.site.host));
  params.set('username',JSON.stringify(identity.username));
  
  return  this.http.get(this.Base+'/stat'+'?'+params.toString(),options)

}


public postMkDir(id: Identity, path: string, name: string ) {
  let headers = new HttpHeaders();
  let options = { headers: headers, withCredentials: true};
  let params = new URLSearchParams;
  params.set('identity',JSON.stringify(id));
  params.set('path',JSON.stringify(path));

  return this.http.post<any>(this.Base+'/mkdir?'+params.toString(),name, options);
}

public getSftpData(id: Identity, path: string, cd: string ) {
  let headers = new HttpHeaders();
  let options = { headers: headers, withCredentials: true};
  let params = new URLSearchParams;
  params.set('identity',JSON.stringify(id));
  params.set('path',JSON.stringify(path));
  params.set('cd',JSON.stringify(cd));

  return this.http.get<any>(this.Base+'/ls?'+params.toString(),options)
                      // .pipe(catchError(this.handleError))

}



private httperror(errorstr: string) {
  var re = /login expired/gi;
  let searchresult = errorstr.search(re);
  if (searchresult != -1) {
    this.notifications.notify("Some authentication tokens have expired, you may need to log in again");
    console.error(errorstr);
    this.authorisationService.updateAgentContents();
  }
  console.error(errorstr);
}

private handleTunnelError(job: Job, error: any) {
  job.connectionState=0;
  if (error.stats == 500) {
    this.notifications.notify(error.error.message);
    return;
  }
  this.notifications.notify('Failed to create a tunnel. Please try again');
  return;
}

private handleAppInstanceError(job: Job, error: any) {
    console.log(error);
    job.connectionState=0;
    if (error.status == 0) {
      this.notifications.notify('It appears we had a timeout learning about your '+job.name+'. Its possible '+job.identity.site.name+' is experiencing issues');
      return;
    }
    if (error.error !== undefined && error.error.message !== undefined) {
        this.notifications.notify(error.error.message);
    } else if (error.error !== undefined && error.error.error !== undefined) {
      this.notifications.notify(error.error.error.msg);
    } else {
        this.notifications.notify(error);
    }
}

private handleError(error: HttpErrorResponse) {
  console.error('in handleError');
  console.error(error);
  if (error.error instanceof ErrorEvent) {
    // A client-side or network error occurred. Handle it accordingly.
    return throwError("Hmm, that didn't work. If you're using a local connection, please make sure Strudel-TES is running.");
  } else {
    // Not sure if this code works correctly. It should update identities in case the error is
    // that the user isn't allowed to run the job

    var re = /identity/gi;
    if (error.error.message != undefined) {
      let searchresult = error.error.message.search(re);
      if (searchresult != -1) {
        // this.getIdentities();
        return throwError('login expired, refreshing');
      }
      return throwError(error.error.message)
    } else {
      return throwError('An error occured but I\'m not sure exactly what. Please try again latter or contact the sys admins');
    }

  }
}





}
